/*!\file: ControlInitialization.cpp
 * \brief: ...
 */ 

#include "../toolkits/toolkits.h"
#include "../objects/objects.h"
#include "../shared/shared.h"
#include "../EnumDefinitions/EnumDefinitions.h"
#include "./parallel.h"
#include "../modules.h"

void ControlInitialization(Model* model){

	extern int my_rank;

	/*fem models: */
	FemModel* fem_dh=NULL;
	FemModel* fem_dv=NULL;
	FemModel* fem_dhu=NULL;
	FemModel* fem_ds=NULL;
	FemModel* fem_sl=NULL;

	/*solutions: */
	Vec ug=NULL;
	Vec ug_horiz=NULL;
	Vec ug_vert=NULL;
	Vec ug_stokes=NULL;
	Vec pg=NULL;
	Vec slopex=NULL;
	Vec slopey=NULL;
	double* vx=NULL;
	double* vy=NULL;
	double* vz=NULL;
	double* pressure=NULL;

	/*flags: */
	int verbose=0;
	int dim=-1;
	int ishutter=0;
	int ismacayealpattyn=0;
	int isstokes=0;
	int numberofdofspernode_sl;
	int numberofdofspernode_dh;
	int numberofdofspernode_ds;
	int numberofnodes;

	double stokesreconditioning;

	/*dof recovery: */
	int dof01[2]={0,1};
	int dof2[1]={2};
	int dof012[3]={0,1,2};
	int dof3[1]={3};
	double* dofset=NULL;

	/*first recover parameters common to all solutions:*/
	model->FindParam(&verbose,VerboseEnum);
	model->FindParam(&dim,DimEnum);
	model->FindParam(&ishutter,IsHutterEnum);
	model->FindParam(&ismacayealpattyn,IsMacAyealPattynEnum);
	model->FindParam(&isstokes,IsStokesEnum);
	model->FindParam(&numberofnodes,NumberOfNodesEnum);
	model->FindParam(&stokesreconditioning,StokesReconditioningEnum);

	/*recover fem models: */
	fem_dh=model->GetFormulation(DiagnosticAnalysisEnum,HorizAnalysisEnum);
	fem_dv=model->GetFormulation(DiagnosticAnalysisEnum,VertAnalysisEnum);
	fem_ds=model->GetFormulation(DiagnosticAnalysisEnum,StokesAnalysisEnum);
	fem_dhu=model->GetFormulation(DiagnosticAnalysisEnum,HutterAnalysisEnum);
	fem_sl=model->GetFormulation(SlopecomputeAnalysisEnum);

	//specific parameters for specific models
	fem_dh->FindParam(&numberofdofspernode_dh,NumberOfDofsPerNodeEnum);
	fem_sl->FindParam(&numberofdofspernode_sl,NumberOfDofsPerNodeEnum);
	fem_ds->FindParam(&numberofdofspernode_ds,NumberOfDofsPerNodeEnum);

	/*if no Stokes, assign output and return*/
	if (!isstokes){
		model->SetActiveFormulation(fem_dh);
		return;
	}

	/*1: compute slopes once for all*/

	//compute slopes
	if(verbose)_printf_("%s\n","computing bed slope (x and y derivatives)...");
	diagnostic_core_linear(&slopex,fem_sl,SlopecomputeAnalysisEnum,BedXAnalysisEnum);
	diagnostic_core_linear(&slopey,fem_sl,SlopecomputeAnalysisEnum,BedYAnalysisEnum);
	FieldExtrudex( slopex, fem_sl->elements,fem_sl->nodes,fem_sl->vertices,fem_sl->loads,fem_sl->materials,fem_sl->parameters,"slopex",0);
	FieldExtrudex( slopey, fem_sl->elements,fem_sl->nodes,fem_sl->vertices,fem_sl->loads,fem_sl->materials,fem_sl->parameters,"slopey",0);

	//Add in inputs
	model->UpdateInputsFromVector(slopex,BedSlopexEnum,VertexEnum);
	model->UpdateInputsFromVector(slopey,BedSlopeyEnum,VertexEnum);
	VecFree(&slopex); VecFree(&slopey);

	/*2: run a complete diagnostic to update spcs*/

	//horizontal velocity
	if(verbose)_printf_("%s\n"," computing horizontal velocities...");
	diagnostic_core_nonlinear(&ug,NULL,NULL,NULL,fem_dh,DiagnosticAnalysisEnum,HorizAnalysisEnum);
	if(verbose)_printf_("%s\n"," extruding horizontal velocities...");
	VecDuplicatePatch(&ug_horiz,ug); FieldExtrudex( ug_horiz,fem_dh->elements,fem_dh->nodes, fem_dh->vertices,fem_dh->loads,fem_dh-> materials,fem_dh->parameters,"velocity",1);

	//Add to inputs:
	SplitSolutionVector(ug_horiz,numberofnodes,numberofdofspernode_dh,&vx,&vy);
	model->UpdateInputsFromVector(vx,VxEnum,VertexEnum);
	model->UpdateInputsFromVector(vy,VyEnum,VertexEnum);
	

	//vertical velocity
	if(verbose)_printf_("%s\n"," computing vertical velocities...");
	diagnostic_core_linear(&ug_vert,fem_dv,DiagnosticAnalysisEnum,VertAnalysisEnum);

	//Create 3d u_g
	if(verbose)_printf_("%s\n"," combining horizontal and vertical velocities...");
	VecFree(&ug); ug=NewVec(numberofnodes*3);
	xfree((void**)&dofset);dofset=dofsetgen(2,&dof01[0],3,numberofnodes*3); VecMerge(ug,ug_horiz,dofset,numberofnodes*2);
	xfree((void**)&dofset);dofset=dofsetgen(1,&dof2[0],3,numberofnodes*3); VecMerge(ug,ug_vert,dofset,numberofnodes*1);
	VecFree(&ug_vert); VecFree(&ug_horiz);

	//Create 4d u_g
	if(verbose)_printf_("%s\n"," computing pressure according to Pattyn...");
	ComputePressurex( &pg,fem_dh->elements, fem_dh->nodes, fem_dh->vertices,fem_dh->loads,  fem_dh->materials, fem_dh->parameters,DiagnosticAnalysisEnum,HorizAnalysisEnum);
	VecScale(pg,1.0/stokesreconditioning);
	ug_stokes=NewVec(fem_ds->nodesets->GetGSize());
	xfree((void**)&dofset);dofset=dofsetgen(3,dof012,4,numberofnodes*4); VecMerge(ug_stokes,ug,dofset,numberofnodes*3);
	xfree((void**)&dofset);dofset=dofsetgen(1,dof3,4,numberofnodes*4); VecMerge(ug_stokes,pg,dofset,numberofnodes);

	//Add in inputs
	SplitSolutionVector(ug_stokes,numberofnodes,numberofdofspernode_ds,&vx,&vy,&vz,&pressure);
	model->UpdateInputsFromVector(vx,VxEnum,VertexEnum);
	model->UpdateInputsFromVector(vy,VyEnum,VertexEnum);
	model->UpdateInputsFromVector(vz,VzEnum,VertexEnum);
	model->UpdateInputsFromVector(pressure,PressureEnum,VertexEnum);
	VecFree(&ug_stokes);

	//update spcs
	if(verbose)_printf_("%s\n"," update boundary conditions for stokes using velocities previously computed...");
	xfree((void**)&dofset);dofset=dofsetgen(3,dof012,4,numberofnodes*4); VecMerge(fem_ds->yg->vector,ug,dofset,3*numberofnodes);
	
	VecFree(&fem_ds->ys); VecFree(&fem_ds->ys0);
	Reducevectorgtosx(&fem_ds->ys,&fem_ds->ys0, fem_ds->yg->vector,fem_ds->nodesets);

	//Compute Stokes velocities to speed up later runs
	if(verbose)_printf_("%s\n"," computing stokes velocities and pressure ...");
	VecFree(&ug);
	diagnostic_core_nonlinear(&ug,NULL,NULL,NULL,fem_ds,DiagnosticAnalysisEnum,StokesAnalysisEnum);

	//Add in inputs
	xfree((void**)&vx); xfree((void**)&vy); xfree((void**)&vz); xfree((void**)&pressure);
	SplitSolutionVector(ug,numberofnodes,numberofdofspernode_ds,&vx,&vy,&vz,&pressure);
	model->UpdateInputsFromVector(vx,VxEnum,VertexEnum);
	model->UpdateInputsFromVector(vy,VyEnum,VertexEnum);
	model->UpdateInputsFromVector(vz,VzEnum,VertexEnum);
	model->UpdateInputsFromVector(pressure,PressureEnum,VertexEnum);

	/*Assign output*/
	model->SetActiveFormulation(fem_ds);

	/*Free ressources:*/
	xfree((void**)&dofset);
	xfree((void**)&vx);
	xfree((void**)&vy);
	xfree((void**)&vz);
	xfree((void**)&pressure);
	VecFree(&ug);
	VecFree(&ug_horiz);
	VecFree(&ug_vert);
	VecFree(&ug_stokes);
	VecFree(&pg);
	VecFree(&slopex);
	VecFree(&slopey);

}
