/*!\file: diagnostic_core.cpp
 * \brief: core of the diagnostic solution 
 */ 

#include "../toolkits/toolkits.h"
#include "../objects/objects.h"
#include "../shared/shared.h"
#include "../EnumDefinitions/EnumDefinitions.h"
#include "./parallel.h"
#include "../modules.h"

void diagnostic_core(DataSet* results,Model* model){

	extern int my_rank;
	int        dummy;

	/*fem models: */
	FemModel* fem_dh=NULL;
	FemModel* fem_dv=NULL;
	FemModel* fem_dhu=NULL;
	FemModel* fem_ds=NULL;
	FemModel* fem_sl=NULL;

	/*output: */
	Result* result=NULL;

	/*solutions: */
	Vec ug=NULL;
	Vec ug_horiz=NULL;
	Vec ug_vert=NULL;
	Vec ug_stokes=NULL;
	Vec pg=NULL;
	Vec riftproperties=NULL;
	double* u_g_initial=NULL;
	double* vx=NULL;
	double* vy=NULL;
	double* vz=NULL;

	/*flags: */
	int verbose=0;
	bool qmu_analysis=0;
	int dim=-1;
	bool ishutter=0;
	bool ismacayealpattyn=0;
	bool isstokes=0;
	int numberofdofspernode_sl;
	int numberofdofspernode_dh;
	int numberofdofspernode_ds;
	int numberofnodes;
	int numrifts=0;

	/*slopes: */
	Vec surfaceslopex=NULL;
	Vec surfaceslopey=NULL;
	Vec bedslopex=NULL;
	Vec bedslopey=NULL;

	double stokesreconditioning;

	/*dof recovery: */
	int dof01[2]={0,1};
	int dof2[1]={2};
	int dof012[3]={0,1,2};
	int dof3[1]={3};
	double* dofset=NULL;


	//first recover parameters common to all solutions
	model->FindParam(&verbose,VerboseEnum);
	model->FindParam(&dim,DimEnum);
	model->FindParam(&ishutter,IsHutterEnum);
	model->FindParam(&ismacayealpattyn,IsMacAyealPattynEnum);
	model->FindParam(&numberofnodes,NumberOfNodesEnum);
	model->FindParam(&isstokes,IsStokesEnum);
	model->FindParam(&stokesreconditioning,StokesReconditioningEnum);
	model->FindParam(&numrifts,NumRiftsEnum);
	model->FindParam(&qmu_analysis,QmuAnalysisEnum);

	/*recover fem models: */
	fem_dh=model->GetFormulation(DiagnosticAnalysisEnum,HorizAnalysisEnum);
	fem_dv=model->GetFormulation(DiagnosticAnalysisEnum,VertAnalysisEnum);
	fem_ds=model->GetFormulation(DiagnosticAnalysisEnum,StokesAnalysisEnum);
	fem_dhu=model->GetFormulation(DiagnosticAnalysisEnum,HutterAnalysisEnum);
	fem_sl=model->GetFormulation(SlopecomputeAnalysisEnum);

	//specific parameters for specific models
	fem_dh->FindParam(&numberofdofspernode_dh,NumberOfDofsPerNodeEnum);
	fem_sl->FindParam(&numberofdofspernode_sl,NumberOfDofsPerNodeEnum);
	fem_ds->FindParam(&numberofdofspernode_ds,NumberOfDofsPerNodeEnum);

	//for qmu analysis, be sure the velocity input we are starting from  is the one in the parameters: */
	if(qmu_analysis){
		model->FindParamByAnalysisAndSub(&vx,&dummy,VxEnum,DiagnosticAnalysisEnum,HorizAnalysisEnum); model->UpdateInputsFromVector(vx,VxEnum,VertexEnum);
		model->FindParamByAnalysisAndSub(&vy,&dummy,VyEnum,DiagnosticAnalysisEnum,HorizAnalysisEnum); model->UpdateInputsFromVector(vy,VyEnum,VertexEnum);
		model->FindParamByAnalysisAndSub(&vz,&dummy,VzEnum,DiagnosticAnalysisEnum,HorizAnalysisEnum); model->UpdateInputsFromVector(vz,VzEnum,VertexEnum);
	}

	/*Compute slopes: */
	slope_core(&surfaceslopex,&surfaceslopey,fem_sl,SurfaceAnalysisEnum);
	slope_core(&bedslopex,&bedslopey,fem_sl,BedAnalysisEnum);
		
	/*Update: */
	model->UpdateInputsFromVector(surfaceslopex,SurfaceSlopexEnum,VertexEnum);
	model->UpdateInputsFromVector(surfaceslopey,SurfaceSlopeyEnum,VertexEnum);
	model->UpdateInputsFromVector(bedslopex,BedSlopexEnum,VertexEnum);
	model->UpdateInputsFromVector(bedslopey,BedSlopeyEnum,VertexEnum);

	
	if(ishutter){
			
		if(verbose)_printf_("%s\n"," computing hutter velocities...");
		diagnostic_core_linear(&ug,fem_dhu,DiagnosticAnalysisEnum,HutterAnalysisEnum);

		if(verbose)_printf_("%s\n"," computing pressure according to MacAyeal...");
		ComputePressurex(&pg,fem_dhu->elements,fem_dhu->nodes, fem_dhu->vertices,fem_dhu->loads,fem_dhu->materials,fem_dhu->parameters,DiagnosticAnalysisEnum,HutterAnalysisEnum);

		if(verbose)_printf_("%s\n"," update boundary conditions for macyeal pattyn using hutter results...");
		if (ismacayealpattyn){
			VecFree(&fem_dh->yg->vector); VecFree(&fem_dh->ys);VecFree(&fem_dh->ys0);
			VecDuplicatePatch(&fem_dh->yg->vector,ug);
			Reducevectorgtosx(&fem_dh->ys,&fem_dh->ys0, fem_dh->yg->vector,fem_dh->nodesets);
		}

	}

	if (ismacayealpattyn){
		
		if(verbose)_printf_("%s\n"," computing horizontal velocities...");
		diagnostic_core_nonlinear(&ug,NULL,NULL,fem_dh->loads,fem_dh,DiagnosticAnalysisEnum,HorizAnalysisEnum);

		if(dim==2){
			if(verbose)_printf_("%s\n"," computing pressure according to MacAyeal...");
			ComputePressurex(&pg,fem_dh->elements,fem_dh->nodes, fem_dh->vertices,fem_dh->loads,fem_dh->materials,fem_dh->parameters,DiagnosticAnalysisEnum,HorizAnalysisEnum);
		}

	}
	
	
	if (dim==3){

		if(verbose)_printf_("%s\n"," extruding horizontal velocities...");
		
		FieldExtrudex( ug,fem_dh->elements,fem_dh->nodes, fem_dh->vertices,fem_dh->loads,fem_dh->materials,fem_dh->parameters,"velocity",1);

		SplitSolutionVector(ug,numberofnodes,numberofdofspernode_dh,&vx,&vy);
		model->UpdateInputsFromVector(vx,VxEnum,VertexEnum);
		model->UpdateInputsFromVector(vy,VyEnum,VertexEnum);
		
		if(verbose)_printf_("%s\n"," computing vertical velocities...");
		diagnostic_core_linear(&ug_vert,fem_dv,DiagnosticAnalysisEnum,VertAnalysisEnum);
		model->UpdateInputsFromVector(ug_vert,VzEnum,VertexEnum);

		if(verbose)_printf_("%s\n"," computing pressure according to Pattyn...");
		ComputePressurex(&pg,fem_dh->elements, fem_dh->nodes, fem_dh->vertices,fem_dh->loads,  fem_dh->materials,fem_dh->parameters,DiagnosticAnalysisEnum,HorizAnalysisEnum);
		model->UpdateInputsFromVector(pg,PressureEnum,VertexEnum);
		
		if (isstokes){

			//"recondition" pressure 
			VecScale(pg,1.0/stokesreconditioning);

			if(verbose)_printf_("%s\n"," update boundary conditions for stokes using velocities previously computed...");
			xfree((void**)&dofset);dofset=dofsetgen(3,dof012,4,numberofnodes*4); VecMerge(fem_ds->yg->vector,ug,dofset,3*numberofnodes);
			VecFree(&fem_ds->ys); VecFree(&fem_ds->ys0);
			Reducevectorgtosx(&fem_ds->ys,&fem_ds->ys0, fem_ds->yg->vector,fem_ds->nodesets);

			if(verbose)_printf_("%s\n"," computing stokes velocities and pressure ...");
			VecFree(&ug);
			diagnostic_core_nonlinear(&ug,NULL,NULL,NULL,fem_ds,DiagnosticAnalysisEnum,StokesAnalysisEnum);
		
			//decondition" pressure
			VecFree(&pg);	
			xfree((void**)&dofset);dofset=dofsetgen(1,dof3,4,numberofnodes*4); VecPartition(&pg, ug, dofset, numberofnodes*1);
			VecScale(pg,stokesreconditioning);
		}
	}
	
	/*Plug results into output dataset: */
	result=new Result(results->Size()+1,0,1,"u_g",ug);
	results->AddObject(result);
	result=new Result(results->Size()+1,0,1,"p_g",pg);
	results->AddObject(result);

	/*output if we have rifts: */
	if(numrifts){
		OutputRiftsx( &riftproperties,fem_dh->loads,numrifts);
		result=new Result(results->Size()+1,0,1,"riftproperties",riftproperties);
		results->AddObject(result);
	}

	/*Free ressources: */
	VecFree(&ug);
	VecFree(&pg);
	xfree((void**)&dofset);
	xfree((void**)&u_g_initial);
	
	VecFree(&surfaceslopex);
	VecFree(&surfaceslopey);
	VecFree(&bedslopex);
	VecFree(&bedslopey);


}
